#! /usr/bin/env python
##########################################################################
# Hopla - Copyright (C) AGrigis, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

# Hopla globals
from .info import DESCRIPTION, __version__  # pragma: no cover
from .scheduler import scheduler  # pragma: no cover
from .converter import hopla  # pragma: no cover

__doc__ = DESCRIPTION  # pragma: no cover
