#! /usr/bin/env python
##########################################################################
# Hopla - Copyright (C) AGrigis, 2015 - 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module defines the scheduler signals.
"""  # pragma: no cover

# Define signals
FLAG_ALL_DONE = b"WORK_FINISHED"  # pragma: no cover
FLAG_WORKER_FINISHED_PROCESSING = (
    b"WORKER_FINISHED_PROCESSING")  # pragma: no cover
